from enum import IntEnum


class DiscoverBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class.  The following are valid Sub Class Id/Sub Class Name value pairs: 0: DBSCHEMA_CATALOGS 1: DBSCHEMA_TABLES 2: DBSCHEMA_COLUMNS 3: DBSCHEMA_PROVIDER_TYPES 4: MDSCHEMA_CUBES 5: MDSCHEMA_DIMENSIONS 6: MDSCHEMA_HIERARCHIES 7: MDSCHEMA_LEVELS 8: MDSCHEMA_MEASURES 9: MDSCHEMA_PROPERTIES 10: MDSCHEMA_MEMBERS 11: MDSCHEMA_FUNCTIONS 12: MDSCHEMA_ACTIONS 13: MDSCHEMA_SETS 14: DISCOVER_INSTANCES 15: MDSCHEMA_KPIS 16: MDSCHEMA_MEASUREGROUPS 17: MDSCHEMA_COMMANDS 18: DMSCHEMA_MINING_SERVICES 19: DMSCHEMA_MINING_SERVICE_PARAMETERS 20: DMSCHEMA_MINING_FUNCTIONS 21: DMSCHEMA_MINING_MODEL_CONTENT 22: DMSCHEMA_MINING_MODEL_XML 23: DMSCHEMA_MINING_MODELS 24: DMSCHEMA_MINING_COLUMNS 25: DISCOVER_DATASOURCES 26: DISCOVER_PROPERTIES 27: DISCOVER_SCHEMA_ROWSETS 28: DISCOVER_ENUMERATORS 29: DISCOVER_KEYWORDS 30: DISCOVER_LITERALS 31: DISCOVER_XML_METADATA 32: DISCOVER_TRACES 33: DISCOVER_TRACE_DEFINITION_PROVIDERINFO 34: DISCOVER_TRACE_COLUMNS 35: DISCOVER_TRACE_EVENT_CATEGORIES 36: DMSCHEMA_MINING_STRUCTURES 37: DMSCHEMA_MINING_STRUCTURE_COLUMNS 38: DISCOVER_MASTER_KEY 39: MDSCHEMA_INPUT_DATASOURCES 40: DISCOVER_LOCATIONS 41: DISCOVER_PARTITION_DIMENSION_STAT 42: DISCOVER_PARTITION_STAT 43: DISCOVER_DIMENSION_STAT 44: MDSCHEMA_MEASUREGROUP_DIMENSIONS 45: DISCOVER_XEVENT_PACKAGES 46: DISCOVER_XEVENT_OBJECTS 47: DISCOVER_XEVENT_OBJECT_COLUMNS 48: DISCOVER_XEVENT_SESSION_TARGETS 49: DISCOVER_XEVENT_SESSIONS 50: DISCOVER_STORAGE_TABLES 51: DISCOVER_STORAGE_TABLE_COLUMNS 52: DISCOVER_STORAGE_TABLE_COLUMN_SEGMENTS 53: DISCOVER_CALC_DEPENDENCY 54: DISCOVER_CSDL_METADATA 55: DISCOVER_RESOURCE_POOLS 56: TMSCHEMA_MODEL 57: TMSCHEMA_DATA_SOURCES 58: TMSCHEMA_TABLES 59: TMSCHEMA_COLUMNS 60: TMSCHEMA_ATTRIBUTE_HIERARCHIES 61: TMSCHEMA_PARTITIONS 62: TMSCHEMA_RELATIONSHIPS 63: TMSCHEMA_MEASURES 64: TMSCHEMA_HIERARCHIES 65: TMSCHEMA_LEVELS 67: TMSCHEMA_TABLE_STORAGES 68: TMSCHEMA_COLUMN_STORAGES 69: TMSCHEMA_PARTITION_STORAGES 70: TMSCHEMA_SEGMENT_MAP_STORAGES 71: TMSCHEMA_DICTIONARY_STORAGES 72: TMSCHEMA_COLUMN_PARTITION_STORAGES 73: TMSCHEMA_RELATIONSHIP_STORAGES 74: TMSCHEMA_RELATIONSHIP_INDEX_STORAGES 75: TMSCHEMA_ATTRIBUTE_HIERARCHY_STORAGES 76: TMSCHEMA_HIERARCHY_STORAGES 77: DISCOVER_RING_BUFFERS 78: TMSCHEMA_KPIS 79: TMSCHEMA_STORAGE_FOLDERS 80: TMSCHEMA_STORAGE_FILES 81: TMSCHEMA_SEGMENT_STORAGES 82: TMSCHEMA_CULTURES 83: TMSCHEMA_OBJECT_TRANSLATIONS 84: TMSCHEMA_LINGUISTIC_METADATA 85: TMSCHEMA_ANNOTATIONS 86: TMSCHEMA_PERSPECTIVES 87: TMSCHEMA_PERSPECTIVE_TABLES 88: TMSCHEMA_PERSPECTIVE_COLUMNS 89: TMSCHEMA_PERSPECTIVE_HIERARCHIES 90: TMSCHEMA_PERSPECTIVE_MEASURES  95: TMSCHEMA_SETS 96: TMSCHEMA_PERSPECTIVE_SETS 97: TMSCHEMA_EXTENDED_PROPERTIES 98: TMSCHEMA_EXPRESSIONS 99: TMSCHEMA_COLUMN_PERMISSIONS 100: TMSCHEMA_DETAIL_ROWS_DEFINITIONS 101: TMSCHEMA_RELATED_COLUMN_DETAILS 102: TMSCHEMA_GROUP_BY_COLUMNS 103: TMSCHEMA_CALCULATION_GROUPS 104: TMSCHEMA_CALCULATION_ITEMS 105: TMSCHEMA_ALTERNATE_OF_DEFINITIONS 106: TMSCHEMA_REFRESH_POLICIES 107: DISCOVER_POWERBI_DATASOURCES 108: TMSCHEMA_FORMAT_STRING_DEFINITIONS 109: DISCOVER_M_EXPRESSIONS 110: TMSCHEMA_POWERBI_ROLES 111: TMSCHEMA_QUERY_GROUPS 112: DISCOVER_DB_MEM_STATS 113: DISCOVER_MEM_STATS 114: TMSCHEMA_ANALYTICS_AIMETADATA 115: DISCOVER_OBJECT_COUNTERS
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the discover event.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the session ID associated with the discover event.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the discover event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the event.
    SERVERNAME = 43  # Contains the name of the instance on which the discover event occurred.
    REQUESTPROPERTIES = (
        45  # Contains the XML for Analysis (XMLA) request properties associated with the discover event.
    )
    ACTIVITYID = 46
    REQUESTID = 47


class DiscoverEndColumns(IntEnum):
    EVENTCLASS = 0  # Contains the event class; this is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class. The following are valid Sub Class Id/Sub Class Name value pairs: 0: DBSCHEMA_CATALOGS 1: DBSCHEMA_TABLES 2: DBSCHEMA_COLUMNS 3: DBSCHEMA_PROVIDER_TYPES 4: MDSCHEMA_CUBES 5: MDSCHEMA_DIMENSIONS 6: MDSCHEMA_HIERARCHIES 7: MDSCHEMA_LEVELS 8: MDSCHEMA_MEASURES 9: MDSCHEMA_PROPERTIES 10: MDSCHEMA_MEMBERS 11: MDSCHEMA_FUNCTIONS 12: MDSCHEMA_ACTIONS 13: MDSCHEMA_SETS 14: DISCOVER_INSTANCES 15: MDSCHEMA_KPIS 16: MDSCHEMA_MEASUREGROUPS 17: MDSCHEMA_COMMANDS 18: DMSCHEMA_MINING_SERVICES 19: DMSCHEMA_MINING_SERVICE_PARAMETERS 20: DMSCHEMA_MINING_FUNCTIONS 21: DMSCHEMA_MINING_MODEL_CONTENT 22: DMSCHEMA_MINING_MODEL_XML 23: DMSCHEMA_MINING_MODELS 24: DMSCHEMA_MINING_COLUMNS 25: DISCOVER_DATASOURCES 26: DISCOVER_PROPERTIES 27: DISCOVER_SCHEMA_ROWSETS 28: DISCOVER_ENUMERATORS 29: DISCOVER_KEYWORDS 30: DISCOVER_LITERALS 31: DISCOVER_XML_METADATA 32: DISCOVER_TRACES 33: DISCOVER_TRACE_DEFINITION_PROVIDERINFO 34: DISCOVER_TRACE_COLUMNS 35: DISCOVER_TRACE_EVENT_CATEGORIES 36: DMSCHEMA_MINING_STRUCTURES 37: DMSCHEMA_MINING_STRUCTURE_COLUMNS 38: DISCOVER_MASTER_KEY 39: MDSCHEMA_INPUT_DATASOURCES 40: DISCOVER_LOCATIONS 41: DISCOVER_PARTITION_DIMENSION_STAT 42: DISCOVER_PARTITION_STAT 43: DISCOVER_DIMENSION_STAT 44: MDSCHEMA_MEASUREGROUP_DIMENSIONS 45: DISCOVER_XEVENT_PACKAGES 46: DISCOVER_XEVENT_OBJECTS 47: DISCOVER_XEVENT_OBJECT_COLUMNS 48: DISCOVER_XEVENT_SESSION_TARGETS 49: DISCOVER_XEVENT_SESSIONS 50: DISCOVER_STORAGE_TABLES 51: DISCOVER_STORAGE_TABLE_COLUMNS 52: DISCOVER_STORAGE_TABLE_COLUMN_SEGMENTS 53: DISCOVER_CALC_DEPENDENCY 54: DISCOVER_CSDL_METADATA 55: DISCOVER_RESOURCE_POOLS 56: TMSCHEMA_MODEL 57: TMSCHEMA_DATA_SOURCES 58: TMSCHEMA_TABLES 59: TMSCHEMA_COLUMNS 60: TMSCHEMA_ATTRIBUTE_HIERARCHIES 61: TMSCHEMA_PARTITIONS 62: TMSCHEMA_RELATIONSHIPS 63: TMSCHEMA_MEASURES 64: TMSCHEMA_HIERARCHIES 65: TMSCHEMA_LEVELS 67: TMSCHEMA_TABLE_STORAGES 68: TMSCHEMA_COLUMN_STORAGES 69: TMSCHEMA_PARTITION_STORAGES 70: TMSCHEMA_SEGMENT_MAP_STORAGES 71: TMSCHEMA_DICTIONARY_STORAGES 72: TMSCHEMA_COLUMN_PARTITION_STORAGES 73: TMSCHEMA_RELATIONSHIP_STORAGES 74: TMSCHEMA_RELATIONSHIP_INDEX_STORAGES 75: TMSCHEMA_ATTRIBUTE_HIERARCHY_STORAGES 76: TMSCHEMA_HIERARCHY_STORAGES 77: DISCOVER_RING_BUFFERS 78: TMSCHEMA_KPIS 79: TMSCHEMA_STORAGE_FOLDERS 80: TMSCHEMA_STORAGE_FILES 81: TMSCHEMA_SEGMENT_STORAGES 82: TMSCHEMA_CULTURES 83: TMSCHEMA_OBJECT_TRANSLATIONS 84: TMSCHEMA_LINGUISTIC_METADATA 85: TMSCHEMA_ANNOTATIONS 86: TMSCHEMA_PERSPECTIVES 87: TMSCHEMA_PERSPECTIVE_TABLES 88: TMSCHEMA_PERSPECTIVE_COLUMNS 89: TMSCHEMA_PERSPECTIVE_HIERARCHIES 90: TMSCHEMA_PERSPECTIVE_MEASURES  95: TMSCHEMA_SETS 96: TMSCHEMA_PERSPECTIVE_SETS 97: TMSCHEMA_EXTENDED_PROPERTIES 98: TMSCHEMA_EXPRESSIONS 99: TMSCHEMA_COLUMN_PERMISSIONS 100: TMSCHEMA_DETAIL_ROWS_DEFINITIONS 101: TMSCHEMA_RELATED_COLUMN_DETAILS 102: TMSCHEMA_GROUP_BY_COLUMNS 103: TMSCHEMA_CALCULATION_GROUPS 104: TMSCHEMA_CALCULATION_ITEMS 105: TMSCHEMA_ALTERNATE_OF_DEFINITIONS 106: TMSCHEMA_REFRESH_POLICIES 107: DISCOVER_POWERBI_DATASOURCES 108: TMSCHEMA_FORMAT_STRING_DEFINITIONS 109: DISCOVER_M_EXPRESSIONS 110: TMSCHEMA_POWERBI_ROLES 111: TMSCHEMA_QUERY_GROUPS 112: DISCOVER_DB_MEM_STATS 113: DISCOVER_MEM_STATS 114: TMSCHEMA_ANALYTICS_AIMETADATA 115: DISCOVER_OBJECT_COUNTERS
    CURRENTTIME = 2  # Contains the current time of the discover event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time (if available) at which the discover end event started.. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Contains the time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Contains the approximate amount of time (milliseconds) taken by the discover event.
    CPUTIME = 6  # Contains the amount of CPU time (in milliseconds) used by the event.
    SEVERITY = 22  # Contains the severity level of an exception.
    SUCCESS = 23  # Contains the success or failure of the discover event. Values are: 0 = Failure 1 = Success
    ERROR = 24  # Contains the error number of any error associated the discover event.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the discover event.
    DATABASENAME = 28  # Contains the name of the database in which the discover event occurred.
    NTUSERNAME = 32  # Contains the Windows user name associated with the object permission event.
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the client process ID of the application that initiated the event.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the session ID associated with the discover event.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the discover end event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the event.
    SERVERNAME = 43  # Contains the name of the instance on which the discover event occurred.
    REQUESTPROPERTIES = 45  # Contains the properties in the XMLA request.
    ACTIVITYID = 46
    REQUESTID = 47



import datetime
from typing import ClassVar
from uuid import UUID, uuid4

from pydantic import PrivateAttr
from structlog import BoundLogger

from pbi_core.logging import get_logger
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.model_tables.enums import DataState, DataType
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

from .commands import CommandMixin
from .enums import Alignment, ColumnType, EncodingHint, SummarizedBy

logger: BoundLogger = get_logger()


class Column(SsasRenameRecord, CommandMixin):  # pyright: ignore[reportIncompatibleMethodOverride]
    """A column of an SSAS table.

    PowerBI spec: [Power BI](https://learn.microsoft.com/en-us/analysis-services/tabular-models/column-properties-ssas-tabular?view=asallproducts-allversions)

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/00a9ec7a-5f4d-4517-8091-b370fe2dc18b)
    """

    _field_mapping: ClassVar[dict[str, str]] = {
        "description": "Description",
    }
    _db_name_field: str = "ExplicitName"
    _read_only_fields = ("table_id",)

    alignment: Alignment
    attribute_hierarchy_id: int
    column_origin_id: int | None = None
    column_storage_id: int
    data_category: str | None = None
    description: str | None = None
    display_folder: str | None = None
    display_ordinal: int
    encoding_hint: EncodingHint
    error_message: str | None = None
    explicit_data_type: DataType  # enum
    explicit_name: str | None = None
    expression: str | int | None = None
    format_string: int | str | None = None
    inferred_data_type: int  # enum
    inferred_name: str | None = None
    is_available_in_mdx: bool
    is_default_image: bool
    is_default_label: bool
    is_hidden: bool
    is_key: bool
    is_nullable: bool
    is_unique: bool
    keep_unique_rows: bool
    lineage_tag: UUID = uuid4()
    sort_by_column_id: int | None = None
    source_column: str | None = None
    state: DataState
    summarize_by: SummarizedBy
    system_flags: int
    table_id: int
    table_detail_position: int
    type: ColumnType

    modified_time: datetime.datetime
    refreshed_time: datetime.datetime
    structure_modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.column)

    def __repr__(self) -> str:
        return f"Column({self.table().name}.{self.pbi_core_name()})"



import datetime

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord, SsasTable
from pbi_core.ssas.model_tables.enums import ObjectType
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import Property


class ObjectTranslation(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/1eade819-5599-4ddd-9bf5-7d365806069d)
    """

    altered: bool
    culture_id: int
    object_id: int
    object_type: ObjectType
    property: Property
    value: str

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.object_translation)

    def object(self) -> SsasTable:
        """Returns the object the annotation is describing.

        Raises:
            TypeError: When the Object Type doesn't map to a know SSAS entity type

        """
        if self.object_type == ObjectType.MODEL:
            return self.tabular_model.model
        mapper = {
            ObjectType.DATASOURCE: self.tabular_model.data_sources,
            ObjectType.TABLE: self.tabular_model.tables,
            ObjectType.COLUMN: self.tabular_model.columns,
            ObjectType.ATTRIBUTE_HIERARCHY: self.tabular_model.attribute_hierarchies,
            ObjectType.PARTITION: self.tabular_model.partitions,
            ObjectType.RELATIONSHIP: self.tabular_model.relationships,
            ObjectType.MEASURE: self.tabular_model.measures,
            ObjectType.HIERARCHY: self.tabular_model.hierarchies,
            ObjectType.LEVEL: self.tabular_model.levels,
            ObjectType.KPI: self.tabular_model.kpis,
            ObjectType.CULTURE: self.tabular_model.cultures,
            ObjectType.LINGUISTIC_METADATA: self.tabular_model.linguistic_metadata,
            ObjectType.PERSPECTIVE: self.tabular_model.perspectives,
            ObjectType.PERSPECTIVE_TABLE: self.tabular_model.perspective_tables,
            ObjectType.PERSPECTIVE_HIERARCHY: self.tabular_model.perspective_hierarchies,
            ObjectType.PERSPECTIVE_MEASURE: self.tabular_model.perspective_measures,
            ObjectType.ROLE: self.tabular_model.roles,
            ObjectType.ROLE_MEMBERSHIP: self.tabular_model.role_memberships,
            ObjectType.TABLE_PERMISSION: self.tabular_model.table_permissions,
            ObjectType.VARIATION: self.tabular_model.variations,
            ObjectType.EXPRESSION: self.tabular_model.expressions,
            ObjectType.COLUMN_PERMISSION: self.tabular_model.column_permissions,
            ObjectType.CALCULATION_GROUP: self.tabular_model.calculation_groups,
            ObjectType.QUERY_GROUP: self.tabular_model.query_groups,
        }
        if self.object_type in mapper:
            return mapper[self.object_type]
        msg = f"No logic implemented for type {self.object_type}"
        raise TypeError(msg)



from enum import Enum, IntEnum
from typing import TYPE_CHECKING, Annotated, Any, cast

import pbi_prototype_query_translation
from pydantic import Discriminator, Tag

from pbi_core.pydantic.main import BaseValidation
from pbi_core.static_files.layout.sources.literal import LiteralSource
from pbi_core.static_files.model_references import ModelColumnReference, ModelMeasureReference

from ._base_node import LayoutNode
from .condition import AndCondition, ComparisonCondition, Condition, ConditionType, InCondition, NotCondition
from .sources import AggregationSource, ColumnSource, Entity, MeasureSource, Source
from .sources.aggregation import ScopedEval
from .visuals.properties.filter_properties import FilterObjects

if TYPE_CHECKING:
    from pbi_prototype_query_translation import TranslationResult

    from pbi_core.ssas.server import LocalTabularModel

    from .bookmark import BookmarkFilters
    from .layout import Layout
    from .section import Section


class Direction(IntEnum):
    ASCENDING = 1
    DESCENDING = 2


class Orderby(LayoutNode):
    Direction: Direction
    Expression: Source


class PrototypeQueryResult(BaseValidation):
    data: list[dict[str, Any]]
    dax_query: str
    column_mapping: dict[str, str]


class InputParameter(LiteralSource):
    Name: str


class InputTableColumn(BaseValidation):
    Expression: Source
    Role: str | None = None


class InputTable(BaseValidation):
    Name: str
    Columns: list[InputTableColumn]


class TransformInput(BaseValidation):
    Parameters: list[InputParameter]
    Table: InputTable


class TransformOutput(BaseValidation):
    Table: InputTable


class TransformMeta(BaseValidation):
    Name: str
    Algorithm: str
    Input: TransformInput
    Output: TransformOutput

    def table_mapping(self) -> dict[str, str]:
        ret: list[ColumnSource | MeasureSource] = []
        for col in self.Input.Table.Columns:
            ret.extend(PrototypeQuery.unwrap_source(col.Expression))
        input_tables: set[str] = set()
        for source in ret:
            if isinstance(source, ColumnSource):
                input_tables.add(source.Column.table())
            else:
                input_tables.add(source.Measure.table())
        if len(input_tables) > 1:
            msg = f"Don't know how to handle multiple inputs: {self}"
            raise ValueError(msg)

        (input_table,) = input_tables
        return {
            self.Output.Table.Name: input_table,
        }


class PrototypeQuery(LayoutNode):
    Version: int
    From: list["From"]
    Select: list[Source] = []
    Where: list[Condition] = []
    OrderBy: list[Orderby] = []
    Transform: list[TransformMeta] = []
    Top: int | None = None

    def table_mapping(self) -> dict[str, str]:
        ret: dict[str, str] = {}
        for from_clause in self.From:
            ret |= from_clause.table_mapping()
        for transform in self.Transform:
            # For measures using Transform outputs, we need to point to the source of that transform table
            transform_tables = transform.table_mapping()
            ret |= {k: ret[v] for k, v in transform_tables.items()}
        return ret

    @classmethod
    def unwrap_source(cls, source: Source | ConditionType | ScopedEval) -> list[ColumnSource | MeasureSource]:
        """Identifies the root sources (measures and columns) used in this filter.

        Raises:
            TypeError: Occurs when one of the source types has not been handled by the code.
                Should not occur outside development.

        """
        if isinstance(source, ColumnSource | MeasureSource):
            return [source]
        if isinstance(source, AggregationSource):
            return cls.unwrap_source(source.Aggregation.Expression)

        if isinstance(source, InCondition):
            ret: list[ColumnSource | MeasureSource] = []
            for expr in source.In.Expressions:
                ret.extend(cls.unwrap_source(expr))
            return ret
        if isinstance(source, NotCondition):
            return cls.unwrap_source(source.Not.Expression)
        if isinstance(source, AndCondition):
            return [
                *cls.unwrap_source(source.And.Left),
                *cls.unwrap_source(source.And.Right),
            ]
        if isinstance(source, ComparisonCondition):
            # Right has no dynamic options, so it's skipped
            return cls.unwrap_source(source.Comparison.Left)
        print(source)
        breakpoint()
        raise TypeError

    def get_ssas_elements(self) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this query is directly dependent on."""
        ret: set[ColumnSource | MeasureSource] = set()
        for select in self.Select:
            ret.update(self.unwrap_source(select))
        for where in self.Where:
            ret.update(self.unwrap_source(where.Condition))
        for order_by in self.OrderBy:
            ret.update(self.unwrap_source(order_by.Expression))
        for transformation in self.Transform:
            for col in transformation.Input.Table.Columns:
                ret.update(self.unwrap_source(col.Expression))
        table_mappings: dict[str, str] = self.table_mapping()
        ret2: set[ModelColumnReference | ModelMeasureReference] = set()
        for source in ret:
            if isinstance(source, ColumnSource):
                ret2.add(
                    ModelColumnReference(
                        column=source.Column.column(),
                        table=table_mappings[source.Column.table()],
                    ),
                )
            else:
                ret2.add(
                    ModelMeasureReference(
                        measure=source.Measure.column(),
                        table=table_mappings[source.Measure.table()],
                    ),
                )
        return ret2

    def get_dax(self, model: "LocalTabularModel") -> "TranslationResult":
        """Creates a DAX query that returns the data for a visual based on the SSAS model supplied.

        Note:
            Although generally the DAX queries generated are identical across different models,
                they can theoretically be different. If you can create a specific case of this,
                please add it to the pbi_core repo!

        Args:
            model (LocalTabularModel): The SSAS model to generate the DAX against.

        Returns:
            DataViewQueryTranslationResult: an object containing the DAX query for this visual

        """
        raw_query = self.model_dump_json()
        return pbi_prototype_query_translation.prototype_query(
            raw_query,
            model.db_name,
            model.server.port,
        )

    def get_data(self, model: "LocalTabularModel") -> PrototypeQueryResult:
        dax_query = self.get_dax(model)
        data = model.server.query_dax(dax_query.dax)

        return PrototypeQueryResult(
            data=data,
            dax_query=dax_query.dax,
            column_mapping=dax_query.column_mapping,
        )


class _SubqueryHelper2(LayoutNode):
    _parent: "_SubqueryHelper"  # pyright: ignore reportIncompatibleVariableOverride=false

    Query: PrototypeQuery


class _SubqueryHelper(LayoutNode):
    _parent: "Subquery"  # pyright: ignore reportIncompatibleVariableOverride=false

    Subquery: _SubqueryHelper2


class SubQueryType(IntEnum):
    NA = 2


class Subquery(LayoutNode):
    _parent: "VisualFilterExpression"  # pyright: ignore reportIncompatibleVariableOverride=false

    Name: str
    Expression: _SubqueryHelper
    Type: SubQueryType

    def table_mapping(self) -> dict[str, str]:
        return self.Expression.Subquery.Query.table_mapping()


def get_from(v: Any) -> str:
    if isinstance(v, dict):
        if "Entity" in v:
            return "Entity"
        if "Expression" in v:
            return "Subquery"
        msg = f"Unknown Filter: {v.keys()}"
        raise ValueError(msg)
    return cast("str", v.__class__.__name__)


From = Annotated[
    Annotated[Entity, Tag("Entity")] | Annotated[Subquery, Tag("Subquery")],
    Discriminator(get_from),
]


class HowCreated(IntEnum):
    AUTO = 0
    """Created automatically when a field is used in the visual."""
    USER = 1
    """Filters created from fields not used in a visual by the user."""
    DRILL = 2
    """Created when drilling down on a data point in a visual."""
    INCLUDE = 3
    """Created by including a data point in a visual."""
    EXCLUDE = 4
    """Created by excluding a data point from a visual."""
    DRILLTHROUGH = 5
    """Created by drill context that is applied to the page when using drill-through
    action from another page."""
    NA3 = 6
    NA4 = 7


class FilterType(Enum):
    ADVANCED = "Advanced"
    CATEGORIAL = "Categorical"
    EXCLUDE = "Exclude"
    INCLUDE = "Include"
    PASSTHROUGH = "Passthrough"
    RANGE = "Range"
    RELATIVE_DATE = "RelativeDate"
    RELATIVE_TIME = "RelativeTime"
    TOP_N = "TopN"
    TUPLE = "Tuple"
    VISUAL_TOP_N = "VisualTopN"


class Scope(LayoutNode):
    scopeId: ConditionType


class CachedDisplayNames(LayoutNode):
    displayName: str
    id: Scope


class Filter(LayoutNode):
    name: str | None = None
    type: FilterType = FilterType.CATEGORIAL
    howCreated: HowCreated = HowCreated.AUTO
    expression: Source | None = None
    """Holds a single expression and associated metadata.
    Name, NativeReferenceName, and Annotations may be specified for any expression.
    Each other property represents a specific type of expression and exactly one of these other properties must be
    specified.

    Definition: https://developer.microsoft.com/json-schemas/fabric/item/report/definition/semanticQuery/1.3.0/schema.json#/definitions/QueryExpressionContainer
    """
    isLockedInViewMode: bool = False
    isHiddenInViewMode: bool = False
    objects: FilterObjects | None = None
    filter: PrototypeQuery | None = None
    """Defines a filter element as a partial query structure"""
    displayName: str | None = None
    ordinal: int = 0
    isLinkedAsAggregation: bool = False
    cachedDisplayNames: list[CachedDisplayNames] | None = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.displayName or self.name})"

    def __str__(self) -> str:
        return super().__str__()

    def get_display_name(self) -> str:
        if self.displayName is not None:
            return self.displayName
        if self.filter is None:
            msg = "Unknown default display name"
            raise ValueError(msg)
        default_name_source = self.filter.Where[0].get_sources()[0]
        if isinstance(default_name_source, ColumnSource):
            return default_name_source.Column.Property
        if isinstance(default_name_source, MeasureSource):
            return default_name_source.Measure.Property
        return "--"

    def get_ssas_elements(self) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this filter is directly dependent on."""
        if self.filter is None:
            return set()
        return self.filter.get_ssas_elements()


class VisualFilterExpression(LayoutNode):
    _parent: "VisualFilter"  # pyright: ignore reportIncompatibleVariableOverride=false

    Version: int | None = None
    From: list["From"] | None = None
    Where: list[Condition]


# TODO: Filter specialization, only done to create better type completion.
# TODO: visual needs extra fields because it allows measure sources I think


class HighlightScope(LayoutNode):
    scopeId: ConditionType


class CachedValueItems(LayoutNode):
    identities: list[HighlightScope]
    valueMap: dict[str, str] | list[str]


class FilterExpressionMetadata(LayoutNode):
    expressions: list[Source]
    cachedValueItems: list[CachedValueItems]


class JsonFilter(LayoutNode):
    filterType: FilterType
    """Type of json filter."""


class DecomposedIdentities(LayoutNode):
    values: list[list[dict[int, Source]]]
    """`values` have 3 levels
    outermost level:
      - for SelectorsByColumn[], it's the number of selectors in this array
      - for FilterExpressionMetadata, it's the number of cachedValueItems.
    second level:
      - for SelectorsByColumn, it is the number of scopedIdentities in the particular SelectorsByColumn
      - for FilterExpressionMetadata, it is the number of identities in a cachedValueItem
    innermost level:
      - the key is the index of the column structure of scopedIdentity in `columns` list;
      - the the value is the expressions list in one identity"""
    columns: list[Source]
    """Defines the set of group-on key columns."""


class DecomposedFilterExpressionMetadata(LayoutNode):
    decomposedIdentities: DecomposedIdentities
    """Defines the group-on key fields and the filters applied on them."""
    expressions: list[Source]
    """Original fields (which have group-on keys) used in the filter."""
    valueMap: list[dict[int, str]] | None = None
    """Matches the index in decomposedIdentities with the queryRef for an expression."""
    jsonFilter: JsonFilter | None = None
    """Json filter metadata."""


def get_filter_expression(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "decomposedIdentities" in v:
            return "DecomposedFilterExpressionMetadata"
        if "cachedValueItems" in v:
            return "FilterExpressionMetadata"
        raise TypeError(v)
    return v.__class__.__name__


FilterExpression = Annotated[
    Annotated[FilterExpressionMetadata, Tag("FilterExpressionMetadata")]
    | Annotated[DecomposedFilterExpressionMetadata, Tag("DecomposedFilterExpressionMetadata")],
    Discriminator(get_filter_expression),
]


class VisualFilter(Filter):
    restatement: str | None = None
    filterExpressionMetadata: FilterExpression | None = None

    def to_bookmark(self) -> "BookmarkFilter":
        return cast("BookmarkFilter", self)


class BookmarkFilter(VisualFilter):
    """Meaning of properties is same as Filters defined outside the bookmark."""

    isTransient: bool = False
    _parent: "BookmarkFilters"  # pyright: ignore reportIncompatibleVariableOverride=false


class PageFilter(Filter):
    _parent: "Section"  # pyright: ignore reportIncompatibleVariableOverride=false


class GlobalFilter(Filter):
    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false



from enum import IntEnum


class FileLoadBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class FileLoadEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    INTEGERDATA = 10  # Integer data.
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class FileSaveBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event
    ACTIVITYID = 46
    REQUESTID = 47


class FileSaveEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    INTEGERDATA = 10  # Integer data.
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class PageOutBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class PageOutEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    INTEGERDATA = 10  # Integer data.
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class PageInBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class PageInEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    JOBID = 7  # Job ID for progress.
    SESSIONTYPE = 8  # Session type (what entity caused the operation).
    INTEGERDATA = 10  # Integer data.
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    SESSIONID = 39  # Session GUID.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47



from ._base_node import LayoutNode
from .layout import Layout
from .resource_package import (
    ResourcePackage,
    ResourcePackageDetails,
    ResourcePackageDetailsType,
    ResourcePackageItem,
    ResourcePackageItemType,
)
from .section import Section

__all__ = [
    "Layout",
    "LayoutNode",
    "ResourcePackage",
    "ResourcePackageDetails",
    "ResourcePackageDetailsType",
    "ResourcePackageItem",
    "ResourcePackageItemType",
    "Section",
]



import json
import pathlib
from typing import Any

import bs4
import jinja2
import requests
from structlog import get_logger  # no need to call the package logger file, since this is just a dev utility

logger = get_logger()

COMMAND_MAPPING = {"create": 0, "alter": 1, "delete": 2, "rename": 3, "refresh": 4}
PREFIX = """
  <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:sql="urn:schemas-microsoft-com:xml-sql">
    <xs:element>
      <xs:complexType>
        <xs:sequence>
          <xs:element type="row"/>
        </xs:sequence>
      </xs:complexType>
    </xs:element>
    <xs:complexType name="row">
      <xs:sequence>
"""[1:-1]  # to remove the leading/trailing newlines
SUFFIX = """
      </xs:sequence>
    </xs:complexType>
  </xs:schema>
"""[1:-1]  # to remove the leading/trailing newlines

command_template = jinja2.Template(
    """
<{{component}}>
{{PREFIX}}
{{fields}}
{{SUFFIX}}
  {% raw %}{{rows}}{% endraw %}
</{{component}}>
""".strip(),
)


BASE_PATH = pathlib.Path(__file__).parent / "command_templates" / "schema"
BASE_PATH.mkdir(exist_ok=True)
BASE_URL = "https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t"


def read_toc_page() -> dict[Any, Any]:
    resp = requests.get(f"{BASE_URL}/toc.json", timeout=10)
    resp_content: dict[Any, Any] = json.loads(resp.text)["items"]
    alter_folder = [0, 0, 0, 0, 0, 2, 0, 4, 1, 0]
    for index in alter_folder:
        resp_content = resp_content[index]["children"]
    return resp_content


def save_entity(entity: dict[str, str], command: str) -> None:
    if entity["toc_title"] in {
        "3.1.5.2.1.5.1.4 Out-of-Line Bindings",
        "3.1.5.2.1.5.1.5 Pushed Data",
    }:
        return
    logger.info(command, entity=entity["toc_title"])
    component = entity["toc_title"].split()[-1]

    page = entity["href"]

    resp = requests.get(f"{BASE_URL}/{page}", timeout=10)
    command_text = bs4.BeautifulSoup(resp.text, "lxml").find("pre")
    if command_text is None:
        raise ValueError
    command_text = command_text.text.replace("Â\xa0", " ")
    command_text = bs4.BeautifulSoup(command_text, "xml")
    row = command_text.find("xs:complexType", {"name": "row"})
    assert isinstance(row, bs4.element.Tag)
    fields = row.find_all("xs:element")
    fields = "\n".join(" " * 8 + str(x) for x in fields)
    (BASE_PATH / component).mkdir(parents=True, exist_ok=True)
    (BASE_PATH / component / f"{command}.xml").write_text(
        command_template.render(PREFIX=PREFIX, fields=fields, SUFFIX=SUFFIX, component=component),
    )


def get_command(command: str, folder: dict[Any, Any]) -> None:
    folder = folder[COMMAND_MAPPING[command]]["children"][0]["children"]
    for page in folder:
        save_entity(page, command)


def main() -> None:
    resp = read_toc_page()
    for command in COMMAND_MAPPING:
        get_command(command, resp)


if __name__ == "__main__":
    main()
